@@ -72,13 +72,11 @@ def train(args, model, device, train_loader, optimizer, privacy_engine, epoch):
         losses.append(loss.item())
 
     if not args.disable_dp:
-        epsilon, best_alpha = privacy_engine.accountant.get_privacy_spent(
-            delta=args.delta
-        )
+        epsilon = privacy_engine.accountant.get_epsilon(delta=args.delta)
         print(
             f"Train Epoch: {epoch} \t"
             f"Loss: {np.mean(losses):.6f} "
-            f"(ε = {epsilon:.2f}, δ = {args.delta}) for α = {best_alpha}"
+            f"(ε = {epsilon:.2f}, δ = {args.delta})"
         )
     else:
         print(f"Train Epoch: {epoch} \t Loss: {np.mean(losses):.6f}")
@@ -224,7 +222,7 @@ def main():
             ),
         ),
         batch_size=args.batch_size,
-        num_workers=1,
+        num_workers=0,
         pin_memory=True,
     )
     test_loader = torch.utils.data.DataLoader(
@@ -240,7 +238,7 @@ def main():
         ),
         batch_size=args.test_batch_size,
         shuffle=True,
-        num_workers=1,
+        num_workers=0,
         pin_memory=True,
     )
     run_results = []
