@@ -14,18 +14,19 @@
 # limitations under the License.
 
 
-from typing import Dict
+from typing import Dict, List
 
 import torch
 import torch.nn as nn
 from opacus.layers.dp_rnn import RNNLinear
+from opt_einsum import contract
 
 from .utils import register_grad_sampler
 
 
 @register_grad_sampler(RNNLinear)
 def compute_rnn_linear_grad_sample(
-    layer: RNNLinear, activations: torch.Tensor, backprops: torch.Tensor
+    layer: RNNLinear, activations: List[torch.Tensor], backprops: torch.Tensor
 ) -> Dict[nn.Parameter, torch.Tensor]:
     """
     Computes per sample gradients for ``RNNLinear`` layer. The RNN-like (DPLSTM, DPGRU) models
@@ -38,11 +39,11 @@ def compute_rnn_linear_grad_sample(
         activations: Activations
         backprops: Backpropagations
     """
-
-    gs = torch.einsum("n...i,n...j->nij", backprops, activations)
-
-    ret = {layer.weight: gs}
-    if layer.bias is not None:
-        ret[layer.bias] = torch.einsum("n...k->nk", backprops)
-
+    activations = activations[0]
+    ret = {}
+    if layer.weight.requires_grad:
+        gs = contract("n...i,n...j->nij", backprops, activations)
+        ret[layer.weight] = gs
+    if layer.bias is not None and layer.bias.requires_grad:
+        ret[layer.bias] = contract("n...k->nk", backprops)
     return ret
