@@ -14,6 +14,7 @@
 # limitations under the License.
 import os
 import warnings
+from itertools import chain
 from typing import IO, Any, BinaryIO, Dict, List, Optional, Tuple, Union
 
 import torch
@@ -21,7 +22,12 @@ from opacus.accountants import create_accountant
 from opacus.accountants.utils import get_noise_multiplier
 from opacus.data_loader import DPDataLoader, switch_generator
 from opacus.distributed import DifferentiallyPrivateDistributedDataParallel as DPDDP
-from opacus.grad_sample.grad_sample_module import GradSampleModule
+from opacus.grad_sample import (
+    AbstractGradSampleModule,
+    GradSampleModule,
+    get_gsm_class,
+    wrap_model,
+)
 from opacus.optimizers import DPOptimizer, get_optimizer_class
 from opacus.scheduler import _NoiseScheduler
 from opacus.utils.module_utils import trainable_parameters
@@ -32,7 +38,9 @@ from torch.utils.data import DataLoader
 
 
 def forbid_accumulation_hook(
-    module: GradSampleModule, _grad_input: torch.Tensor, _grad_output: torch.Tensor
+    module: AbstractGradSampleModule,
+    _grad_input: torch.Tensor,
+    _grad_output: torch.Tensor,
 ):
     """
     Model hook that detects repetitive forward/backward passes between optimizer steps.
@@ -99,13 +107,14 @@ class PrivacyEngine:
         >>> # continue training as normal
     """
 
-    def __init__(self, *, accountant: str = "rdp", secure_mode: bool = False):
+    def __init__(self, *, accountant: str = "prv", secure_mode: bool = False):
         """
 
         Args:
             accountant: Accounting mechanism. Currently supported:
                 - rdp (:class:`~opacus.accountants.RDPAccountant`)
                 - gdp (:class:`~opacus.accountants.GaussianAccountant`)
+                - prv (:class`~opacus.accountants.PRVAccountant`)
             secure_mode: Set to ``True`` if cryptographically strong DP guarantee is
                 required. ``secure_mode=True`` uses secure random number generator for
                 noise and shuffling (as opposed to pseudo-rng in vanilla PyTorch) and
@@ -147,6 +156,7 @@ class PrivacyEngine:
         distributed: bool = False,
         clipping: str = "flat",
         noise_generator=None,
+        grad_sample_mode="hooks",
     ) -> DPOptimizer:
         if isinstance(optimizer, DPOptimizer):
             optimizer = optimizer.original_optimizer
@@ -157,7 +167,11 @@ class PrivacyEngine:
         elif noise_generator is not None:
             generator = noise_generator
 
-        optim_class = get_optimizer_class(clipping=clipping, distributed=distributed)
+        optim_class = get_optimizer_class(
+            clipping=clipping,
+            distributed=distributed,
+            grad_sample_mode=grad_sample_mode,
+        )
 
         return optim_class(
             optimizer=optimizer,
@@ -203,28 +217,33 @@ class PrivacyEngine:
         *,
         batch_first: bool = True,
         loss_reduction: str = "mean",
-    ) -> GradSampleModule:
+        grad_sample_mode: str = "hooks",
+    ) -> AbstractGradSampleModule:
         # Ideally, validation should have been taken care of by calling
         # `get_compatible_module()`
         self.validate(module=module, optimizer=None, data_loader=None)
 
         # wrap
-        if isinstance(module, GradSampleModule):
+        if isinstance(module, AbstractGradSampleModule):
             if (
                 module.batch_first != batch_first
                 or module.loss_reduction != loss_reduction
+                or type(module) != get_gsm_class(grad_sample_mode)
             ):
                 raise ValueError(
                     f"Pre-existing GradSampleModule doesn't match new arguments."
-                    f"Got: module.batch_first: {module.batch_first}, module.loss_reduction: {module.loss_reduction}"
-                    f"Requested: batch_first:{batch_first}, loss_reduction: {loss_reduction}. "
+                    f"Got: module.batch_first: {module.batch_first}, module.loss_reduction: {module.loss_reduction}, type(module): {type(module)}"
+                    f"Requested: batch_first:{batch_first}, loss_reduction: {loss_reduction}, grad_sample_mode: {grad_sample_mode} "
                     f"Please pass vanilla nn.Module instead"
                 )
 
             return module
         else:
-            return GradSampleModule(
-                module, batch_first=batch_first, loss_reduction=loss_reduction
+            return wrap_model(
+                module,
+                grad_sample_mode=grad_sample_mode,
+                batch_first=batch_first,
+                loss_reduction=loss_reduction,
             )
 
     def is_compatible(
@@ -301,9 +320,10 @@ class PrivacyEngine:
         poisson_sampling: bool = True,
         clipping: str = "flat",
         noise_generator=None,
+        grad_sample_mode: str = "hooks",
     ) -> Tuple[GradSampleModule, DPOptimizer, DataLoader]:
         """
-        Add privacy-related responsibilites to the main PyTorch training objects:
+        Add privacy-related responsibilities to the main PyTorch training objects:
         model, optimizer, and the data loader.
 
         All of the returned objects act just like their non-private counterparts
@@ -345,6 +365,10 @@ class PrivacyEngine:
                 with distributed training can provide notable performance gains.
             noise_generator: torch.Generator() object used as a source of randomness for
                 the noise
+            grad_sample_mode: mode for computing per sample gradients. Determines the
+                implementation class for the wrapped ``module``. See
+                :class:`~opacus.grad_sample.gsm_base.AbstractGradSampleModule` for more
+                details
 
         Returns:
             Tuple of (model, optimizer, data_loader).
@@ -361,24 +385,22 @@ class PrivacyEngine:
             raise ValueError("Passing seed is prohibited in secure mode")
 
         # compare module parameter with optimizer parameters
-        if not all(
-            torch.eq(i, j).all()
-            for i, j in zip(
-                list(module.parameters()),
-                sum(
-                    [param_group["params"] for param_group in optimizer.param_groups],
-                    [],
-                ),
-            )
+        model_parameters = set(module.parameters())
+        for p in chain.from_iterable(
+            [param_group["params"] for param_group in optimizer.param_groups]
         ):
-            raise ValueError(
-                "Module parameters are different than optimizer Parameters"
-            )
+            if p not in model_parameters:
+                raise ValueError(
+                    "Module parameters are different than optimizer Parameters"
+                )
 
         distributed = isinstance(module, (DPDDP, DDP))
 
         module = self._prepare_model(
-            module, batch_first=batch_first, loss_reduction=loss_reduction
+            module,
+            batch_first=batch_first,
+            loss_reduction=loss_reduction,
+            grad_sample_mode=grad_sample_mode,
         )
         if poisson_sampling:
             module.register_backward_hook(forbid_accumulation_hook)
@@ -404,6 +426,7 @@ class PrivacyEngine:
             noise_generator=noise_generator,
             distributed=distributed,
             clipping=clipping,
+            grad_sample_mode=grad_sample_mode,
         )
 
         optimizer.attach_step_hook(
@@ -421,10 +444,13 @@ class PrivacyEngine:
         target_epsilon: float,
         target_delta: float,
         epochs: int,
-        max_grad_norm: float,
+        max_grad_norm: Union[float, List[float]],
         batch_first: bool = True,
         loss_reduction: str = "mean",
+        poisson_sampling: bool = True,
+        clipping: str = "flat",
         noise_generator=None,
+        grad_sample_mode: str = "hooks",
         **kwargs,
     ):
         """
@@ -438,9 +464,6 @@ class PrivacyEngine:
             module: PyTorch module to be used for training
             optimizer: Optimizer to be used for training
             data_loader: DataLoader to be used for training
-            noise_multiplier: The ratio of the standard deviation of the Gaussian noise to
-                the L2-sensitivity of the function to which the noise is added
-                (How much noise to add)
             max_grad_norm: The maximum norm of the per-sample gradients. Any gradient with norm
                 higher than this will be clipped to this value.
             batch_first: Flag to indicate if the input tensor to the corresponding module
@@ -449,18 +472,23 @@ class PrivacyEngine:
                 ``[K, batch_size, ...]``
             loss_reduction: Indicates if the loss reduction (for aggregating the gradients)
                 is a sum or a mean operation. Can take values "sum" or "mean"
-            noise_seed: Seed to be used for random noise generation
             poisson_sampling: ``True`` if you want to use standard sampling required
                 for DP guarantees. Setting ``False`` will leave provided data_loader
                 unchanged. Technically this doesn't fit the assumptions made by
                 privacy accounting mechanism, but it can be a good approximation when
                 using Poisson sampling is unfeasible.
-            clipping: Per sample gradient clipping mechanism ("flat" or "per_layer").
+            clipping: Per sample gradient clipping mechanism ("flat" or "per_layer" or "adaptive").
                 Flat clipping calculates the norm of the entire gradient over
                 all parameters, per layer clipping sets individual norms for
                 every parameter tensor, and adaptive clipping updates clipping bound per iteration.
                 Flat clipping is usually preferred, but using per layer clipping in combination
                 with distributed training can provide notable performance gains.
+            noise_generator: torch.Generator() object used as a source of randomness for
+                the noise
+            grad_sample_mode: mode for computing per sample gradients. Determines the
+                implementation class for the wrapped ``module``. See
+                :class:`~opacus.grad_sample.gsm_base.AbstractGradSampleModule` for more
+                details
 
         Returns:
             Tuple of (model, optimizer, data_loader).
@@ -468,7 +496,7 @@ class PrivacyEngine:
             Model is a wrapper around the original model that also computes per sample
                 gradients
             Optimizer is a wrapper around the original optimizer that also does
-             gradient clipping and adding noise to the gradients
+                gradient clipping and noise addition to the gradients
             DataLoader is a brand new DataLoader object, constructed to behave as
                 equivalent to the original data loader, possibly with updated
                 sampling mechanism. Points to the same dataset object.
@@ -498,6 +526,9 @@ class PrivacyEngine:
             batch_first=batch_first,
             loss_reduction=loss_reduction,
             noise_generator=noise_generator,
+            grad_sample_mode=grad_sample_mode,
+            poisson_sampling=poisson_sampling,
+            clipping=clipping,
         )
 
     def get_epsilon(self, delta):
@@ -524,7 +555,7 @@ class PrivacyEngine:
         torch_save_kwargs: Optional[Dict[str, Any]] = None,
     ):
         """
-        Saves the state_dict of module, optimzer, and accountant at path.
+        Saves the state_dict of module, optimizer, and accountant at path.
         Args:
             path: Path to save the state dict objects.
             module: GradSampleModule to save; wrapped module's state_dict is saved.
