@@ -19,9 +19,13 @@ import torch
 import torch.nn as nn
 import torch.nn.functional as F
 from opacus.grad_sample import GradSampleModule
+from opacus.grad_sample.gsm_exp_weights import (
+    API_CUTOFF_VERSION,
+    GradSampleModuleExpandedWeights,
+)
 from opacus.grad_sample.linear import compute_linear_grad_sample
 from opacus.grad_sample.utils import register_grad_sampler
-from torch.testing import assert_allclose
+from torch.testing import assert_close
 from torch.utils.data import DataLoader
 from torchvision import transforms
 from torchvision.datasets import FakeData
@@ -52,6 +56,9 @@ class SampleConvNet(nn.Module):
 
 
 class GradSampleModuleTest(unittest.TestCase):
+
+    CLS = GradSampleModule
+
     def setUp(self):
         self.original_model = SampleConvNet()
         copy_of_original_model = SampleConvNet()
@@ -59,7 +66,7 @@ class GradSampleModuleTest(unittest.TestCase):
             self.original_model.state_dict(), strict=True
         )
 
-        self.grad_sample_module = GradSampleModule(
+        self.grad_sample_module = self.CLS(
             copy_of_original_model, batch_first=True, loss_reduction="mean"
         )
         self.DATA_SIZE = 8
@@ -99,7 +106,7 @@ class GradSampleModuleTest(unittest.TestCase):
             f"MSE = {F.mse_loss(gs_out, normal_out)}, ",
             f"L1 Loss = {F.l1_loss(gs_out, normal_out)}",
         )
-        assert_allclose(gs_out, normal_out, atol=1e-7, rtol=1e-5, msg=msg)
+        assert_close(gs_out, normal_out, atol=1e-7, rtol=1e-5, msg=msg)
 
     def test_zero_grad(self):
         x, _ = next(iter(self.dl))
@@ -154,7 +161,7 @@ class GradSampleModuleTest(unittest.TestCase):
                 f"L1 Loss = {F.l1_loss(gs_tensor, original_tensor)}",
             )
 
-            assert_allclose(gs_tensor, original_tensor, atol=1e-6, rtol=1e-4, msg=msg)
+            assert_close(gs_tensor, original_tensor, atol=1e-6, rtol=1e-4, msg=msg)
 
     def test_remove_hooks(self):
         """
@@ -205,11 +212,18 @@ class GradSampleModuleTest(unittest.TestCase):
             def forward(self, x: torch.Tensor):
                 return F.linear(x, self.p)
 
-        with self.assertRaises(NotImplementedError):
-            GradSampleModule(SimpleLinear(4, 2))
+        # Should be handled by functorch
+        try:
+            gsm = GradSampleModule(SimpleLinear(4, 2))
+            self.assertTrue(hasattr(gsm._module, "ft_compute_sample_grad"))
+        except ImportError:
+            print("Test could not be ran because functorch not available")
 
         # Should not raise exception if strict=False
-        GradSampleModule(SimpleLinear(4, 2), strict=False)
+        try:
+            GradSampleModule(SimpleLinear(4, 2), strict=False)
+        except ImportError:
+            print("Test could not be ran because functorch not available")
 
         # Should not fail after relevant grad sampler has been registered
         register_grad_sampler(SimpleLinear)(compute_linear_grad_sample)
@@ -219,9 +233,6 @@ class GradSampleModuleTest(unittest.TestCase):
         with self.assertRaises(NotImplementedError):
             GradSampleModule(mobilenet_v3_small())
 
-        # Should not raise exception if strict=False
-        GradSampleModule(mobilenet_v3_small(), strict=False)
-
     def test_submodule_access(self):
         _ = self.grad_sample_module.fc1
         _ = self.grad_sample_module.fc2
@@ -235,7 +246,7 @@ class GradSampleModuleTest(unittest.TestCase):
         # check wrapped module state dict
         for key in og_state_dict.keys():
             self.assertTrue(f"_module.{key}" in gs_state_dict)
-            assert_allclose(og_state_dict[key], gs_state_dict[f"_module.{key}"])
+            assert_close(og_state_dict[key], gs_state_dict[f"_module.{key}"])
 
     def test_load_state_dict(self):
         gs_state_dict = self.grad_sample_module.state_dict()
@@ -246,6 +257,22 @@ class GradSampleModuleTest(unittest.TestCase):
         # wrapped module is the same
         for key in self.original_model.state_dict().keys():
             self.assertTrue(key in new_gs._module.state_dict())
-            assert_allclose(
+            assert_close(
                 self.original_model.state_dict()[key], new_gs._module.state_dict()[key]
             )
+
+
+@unittest.skipIf(
+    torch.__version__ < API_CUTOFF_VERSION, "not supported in this torch version"
+)
+class EWGradSampleModuleTest(GradSampleModuleTest):
+    CLS = GradSampleModuleExpandedWeights
+
+    def test_remove_hooks(self):
+        pass
+
+    def test_enable_hooks(self):
+        pass
+
+    def test_disable_hooks(self):
+        pass
