@@ -15,7 +15,7 @@
 
 import io
 import unittest
-from typing import Dict, List, Union
+from typing import Dict, Iterable, List, Tuple, Union
 
 import numpy as np
 import torch
@@ -25,7 +25,7 @@ from opacus.grad_sample import wrap_model
 from opacus.utils.module_utils import trainable_parameters
 from opacus.utils.packed_sequences import compute_seq_lengths
 from torch.nn.utils.rnn import PackedSequence, pad_packed_sequence
-from torch.testing import assert_allclose
+from torch.testing import assert_close
 
 
 def expander(x, factor: int = 2):
@@ -36,6 +36,13 @@ def shrinker(x, factor: int = 2):
     return max(1, x // factor)  # if avoid returning 0 for x == 1
 
 
+def is_batch_empty(batch: Union[torch.Tensor, Iterable[torch.Tensor]]):
+    if type(batch) is torch.Tensor:
+        return batch.numel() == 0
+    else:
+        return batch[0].numel() == 0
+
+
 class ModelWithLoss(nn.Module):
     """
     To test the gradients of a module, we need to have a loss.
@@ -67,7 +74,10 @@ class ModelWithLoss(nn.Module):
         self.criterion = nn.L1Loss(reduction=loss_reduction)
 
     def forward(self, x):
-        x = self.wrapped_module(x)
+        if type(x) is tuple:
+            x = self.wrapped_module(*x)
+        else:
+            x = self.wrapped_module(x)
         if type(x) is PackedSequence:
             loss = _compute_loss_packedsequences(self.criterion, x)
         else:
@@ -106,6 +116,7 @@ class GradSampleHooks_test(unittest.TestCase):
         module: nn.Module,
         batch_first=True,
         loss_reduction="mean",
+        chunk_method=iter,
     ) -> Dict[str, torch.tensor]:
         """
         Computes per-sample gradients with the microbatch method, i.e. by computing normal gradients
@@ -117,6 +128,8 @@ class GradSampleHooks_test(unittest.TestCase):
             module: The ``ModelWithLoss`` that wraps the nn.Module you want to test.
             batch_first: Whether batch size is the first dimension (as opposed to the second).
                 Defaults to True.
+            loss_reduction: What reduction to apply to the loss. Defaults to "mean".
+            chunk_method: The method to use to split the batch into microbatches. Defaults to ``iter``.
 
         Returns:
             Dictionary mapping parameter_name -> per-sample-gradient for that parameter
@@ -136,12 +149,14 @@ class GradSampleHooks_test(unittest.TestCase):
 
         # Invariant: x is [B, T, ...]
 
-        for x_i in x:
+        for x_i in chunk_method(x):
             # x_i is [T, ...]
-            x_i = x_i.unsqueeze(
-                0 if batch_first else 1
-            )  # x_i of size [1, T, ...] if batch_first, else [T, 1, ...]
             module.zero_grad()
+            if type(x_i) is not tuple:
+                # EmbeddingBag provides tuples
+                x_i = x_i.unsqueeze(
+                    0 if batch_first else 1
+                )  # x_i of size [1, T, ...] if batch_first, else [T, 1, ...]
             loss_i = module(x_i)
             loss_i.backward()
             for p in module.parameters():
@@ -213,12 +228,13 @@ class GradSampleHooks_test(unittest.TestCase):
 
     def run_test(
         self,
-        x: Union[torch.Tensor, PackedSequence],
+        x: Union[torch.Tensor, PackedSequence, Tuple],
         module: nn.Module,
         batch_first=True,
         atol=10e-6,
         rtol=10e-5,
         ew_compatible=True,
+        chunk_method=iter,
     ):
         grad_sample_modes = ["hooks", "functorch"]
         try:
@@ -226,6 +242,11 @@ class GradSampleHooks_test(unittest.TestCase):
         except ImportError:
             grad_sample_modes = ["hooks"]
 
+        if type(module) is nn.EmbeddingBag or (
+            type(x) is not PackedSequence and is_batch_empty(x)
+        ):
+            grad_sample_modes = ["hooks"]
+
         for grad_sample_mode in grad_sample_modes:
             for loss_reduction in ["sum", "mean"]:
 
@@ -240,6 +261,7 @@ class GradSampleHooks_test(unittest.TestCase):
                         atol=atol,
                         rtol=rtol,
                         grad_sample_mode=grad_sample_mode,
+                        chunk_method=chunk_method,
                     )
         if ew_compatible and batch_first and torch.__version__ >= (1, 13):
             self.run_test_with_reduction(
@@ -250,6 +272,7 @@ class GradSampleHooks_test(unittest.TestCase):
                 atol=atol,
                 rtol=rtol,
                 grad_sample_mode="ew",
+                chunk_method=chunk_method,
             )
 
     def run_test_with_reduction(
@@ -261,7 +284,16 @@ class GradSampleHooks_test(unittest.TestCase):
         atol=10e-6,
         rtol=10e-5,
         grad_sample_mode="hooks",
+        chunk_method=iter,
     ):
+        opacus_grad_samples = self.compute_opacus_grad_sample(
+            x,
+            module,
+            batch_first=batch_first,
+            loss_reduction=loss_reduction,
+            grad_sample_mode=grad_sample_mode,
+        )
+
         if type(x) is PackedSequence:
             x_unpacked = _unpack_packedsequences(x)
             microbatch_grad_samples = self.compute_microbatch_grad_sample(
@@ -270,18 +302,17 @@ class GradSampleHooks_test(unittest.TestCase):
                 batch_first=batch_first,
                 loss_reduction=loss_reduction,
             )
-        else:
+        elif not is_batch_empty(x):
             microbatch_grad_samples = self.compute_microbatch_grad_sample(
-                x, module, batch_first=batch_first, loss_reduction=loss_reduction
+                x,
+                module,
+                batch_first=batch_first,
+                loss_reduction=loss_reduction,
+                chunk_method=chunk_method,
             )
-
-        opacus_grad_samples = self.compute_opacus_grad_sample(
-            x,
-            module,
-            batch_first=batch_first,
-            loss_reduction=loss_reduction,
-            grad_sample_mode=grad_sample_mode,
-        )
+        else:
+            # We've checked opacus can handle 0-sized batch. Microbatch doesn't make sense
+            return
 
         if microbatch_grad_samples.keys() != opacus_grad_samples.keys():
             raise ValueError(
@@ -344,7 +375,7 @@ class GradSampleHooks_test(unittest.TestCase):
                 f"L1 Loss = {F.l1_loss(opacus_grad_sample, microbatch_grad_sample)}",
             )
             try:
-                assert_allclose(
+                assert_close(
                     actual=microbatch_grad_sample,
                     expected=opacus_grad_sample,
                     atol=atol,
