@@ -25,7 +25,7 @@ from .common import GradSampleHooks_test, expander, shrinker
 
 class Conv3d_test(GradSampleHooks_test):
     @given(
-        N=st.integers(1, 4),
+        N=st.integers(0, 4),
         C=st.sampled_from([1, 3, 32]),
         D=st.integers(3, 6),
         H=st.integers(6, 10),
@@ -33,11 +33,11 @@ class Conv3d_test(GradSampleHooks_test):
         out_channels_mapper=st.sampled_from([expander, shrinker]),
         kernel_size=st.sampled_from([2, 3, (1, 2, 3)]),
         stride=st.sampled_from([1, 2, (1, 2, 3)]),
-        padding=st.sampled_from([0, 2, (1, 2, 3)]),
+        padding=st.sampled_from([0, 2, (1, 2, 3), "same", "valid"]),
         dilation=st.sampled_from([1, (1, 2, 2)]),
         groups=st.integers(1, 16),
     )
-    @settings(deadline=10000)
+    @settings(deadline=30000)
     def test_conv3d(
         self,
         N: int,
@@ -53,6 +53,8 @@ class Conv3d_test(GradSampleHooks_test):
         groups: int,
     ):
 
+        if padding == "same" and stride != 1:
+            return
         out_channels = out_channels_mapper(C)
         if (
             C % groups != 0 or out_channels % groups != 0
@@ -68,4 +70,14 @@ class Conv3d_test(GradSampleHooks_test):
             dilation=dilation,
             groups=groups,
         )
-        self.run_test(x, conv, batch_first=True, atol=10e-5, rtol=10e-3)
+        is_ew_compatible = (
+            dilation == 1 and padding != "same" and N > 0
+        )  # TODO add support for padding = 'same' with EW
+        self.run_test(
+            x,
+            conv,
+            batch_first=True,
+            atol=10e-5,
+            rtol=10e-3,
+            ew_compatible=is_ew_compatible,
+        )
