@@ -75,6 +75,7 @@ class RNN_test(GradSampleHooks_test):
         using_packed_sequences: bool,
         packed_sequences_sorted: bool,
     ):
+        torch.use_deterministic_algorithms(False)
         rnn = model(
             D,
             H,
@@ -92,4 +93,4 @@ class RNN_test(GradSampleHooks_test):
                 x = torch.randn([N, T, D])
             else:
                 x = torch.randn([T, N, D])
-        self.run_test(x, rnn, batch_first=batch_first)
+        self.run_test(x, rnn, batch_first=batch_first, ew_compatible=False)
