@@ -1,5 +1,17 @@
 #!/usr/bin/env python3
-# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
+# Copyright (c) Meta Platforms, Inc. and affiliates.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 
 import hypothesis.strategies as st
 import torch
@@ -11,7 +23,7 @@ from .common import GradSampleHooks_test
 
 class Embedding_test(GradSampleHooks_test):
     @given(
-        N=st.integers(1, 4),
+        N=st.integers(0, 4),
         T=st.integers(1, 5),
         Q=st.integers(1, 4),
         R=st.integers(1, 2),
@@ -43,5 +55,5 @@ class Embedding_test(GradSampleHooks_test):
             size = [N, T, Q, R]
 
         emb = nn.Embedding(V, D)
-        x = torch.randint(low=0, high=V - 1, size=size)
-        self.run_test(x, emb, batch_first=batch_first)
+        x = torch.randint(low=0, high=V, size=size)
+        self.run_test(x, emb, batch_first=batch_first, ew_compatible=N > 0)
