@@ -1,5 +1,17 @@
 #!/usr/bin/env python3
-# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
+# Copyright (c) Meta Platforms, Inc. and affiliates.
+#
+# Licensed under the Apache License, Version 2.0 (the "License");
+# you may not use this file except in compliance with the License.
+# You may obtain a copy of the License at
+#
+#     http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing, software
+# distributed under the License is distributed on an "AS IS" BASIS,
+# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
+# See the License for the specific language governing permissions and
+# limitations under the License.
 
 from typing import Union
 
@@ -18,7 +30,7 @@ class GroupNorm_test(GradSampleHooks_test):
     """
 
     @given(
-        N=st.integers(1, 4),
+        N=st.integers(0, 4),
         C=st.integers(1, 8),
         H=st.integers(5, 10),
         W=st.integers(4, 8),
@@ -42,4 +54,4 @@ class GroupNorm_test(GradSampleHooks_test):
 
         x = torch.randn([N, C, H, W])
         norm = nn.GroupNorm(num_groups=num_groups, num_channels=C, affine=True)
-        self.run_test(x, norm, batch_first=True)
+        self.run_test(x, norm, batch_first=True, ew_compatible=N > 0)
