@@ -15,10 +15,11 @@
 
 import abc
 import io
+import itertools
 import math
 import unittest
 from abc import ABC
-from typing import Optional, OrderedDict
+from typing import Optional, OrderedDict, Type
 from unittest.mock import MagicMock, patch
 
 import hypothesis.strategies as st
@@ -27,9 +28,10 @@ import torch.nn as nn
 import torch.nn.functional as F
 from hypothesis import given, settings
 from opacus import PrivacyEngine
+from opacus.grad_sample.gsm_exp_weights import API_CUTOFF_VERSION
 from opacus.layers.dp_multihead_attention import DPMultiheadAttention
 from opacus.optimizers.optimizer import _generate_noise
-from opacus.scheduler import StepNoise, _NoiseScheduler
+from opacus.scheduler import StepNoise
 from opacus.utils.module_utils import are_state_dict_equal
 from opacus.validators.errors import UnsupportedModuleError
 from opacus.validators.module_validator import ModuleValidator
@@ -38,6 +40,18 @@ from torch.utils.data import DataLoader, Dataset, TensorDataset
 from torchvision import models, transforms
 from torchvision.datasets import FakeData
 
+from .utils import CustomLinearModule, LinearWithExtraParam
+
+
+def _is_functorch_available():
+    try:
+        # flake8: noqa F401
+        import functorch
+
+        return True
+    except ImportError:
+        return False
+
 
 def get_grad_sample_aggregated(tensor: torch.Tensor, loss_type: str = "mean"):
     if tensor.grad_sample is None:
@@ -65,6 +79,7 @@ class BasePrivacyEngineTest(ABC):
         self.ALPHAS = [1 + x / 10.0 for x in range(1, 100, 10)]
         self.criterion = nn.CrossEntropyLoss()
         self.BATCH_FIRST = True
+        self.GRAD_SAMPLE_MODE = "hooks"
 
         torch.manual_seed(42)
 
@@ -79,9 +94,16 @@ class BasePrivacyEngineTest(ABC):
     def _init_vanilla_training(
         self,
         state_dict: Optional[OrderedDict[str, torch.Tensor]] = None,
+        opt_exclude_frozen=False,
     ):
         model = self._init_model()
-        optimizer = torch.optim.SGD(model.parameters(), lr=self.LR, momentum=0)
+        optimizer = torch.optim.SGD(
+            model.parameters()
+            if not opt_exclude_frozen
+            else [p for p in model.parameters() if p.requires_grad],
+            lr=self.LR,
+            momentum=0,
+        )
         if state_dict:
             model.load_state_dict(state_dict)
         dl = self._init_data()
@@ -95,10 +117,18 @@ class BasePrivacyEngineTest(ABC):
         max_grad_norm: float = 1.0,
         poisson_sampling: bool = True,
         clipping: str = "flat",
+        grad_sample_mode="hooks",
+        opt_exclude_frozen=False,
     ):
         model = self._init_model()
         model = PrivacyEngine.get_compatible_module(model)
-        optimizer = torch.optim.SGD(model.parameters(), lr=self.LR, momentum=0)
+        optimizer = torch.optim.SGD(
+            model.parameters()
+            if not opt_exclude_frozen
+            else [p for p in model.parameters() if p.requires_grad],
+            lr=self.LR,
+            momentum=0,
+        )
 
         if state_dict:
             model.load_state_dict(state_dict)
@@ -119,6 +149,7 @@ class BasePrivacyEngineTest(ABC):
             batch_first=self.BATCH_FIRST,
             poisson_sampling=poisson_sampling,
             clipping=clipping,
+            grad_sample_mode=grad_sample_mode,
         )
 
         return model, optimizer, poisson_dl, privacy_engine
@@ -175,20 +206,34 @@ class BasePrivacyEngineTest(ABC):
                     break
 
     def test_basic(self):
-        model, optimizer, dl, _ = self._init_private_training(
-            noise_multiplier=1.0,
-            max_grad_norm=1.0,
-            poisson_sampling=True,
-        )
-        self._train_steps(model, optimizer, dl)
+        for opt_exclude_frozen in [True, False]:
+            with self.subTest(opt_exclude_frozen=opt_exclude_frozen):
+                model, optimizer, dl, _ = self._init_private_training(
+                    noise_multiplier=1.0,
+                    max_grad_norm=1.0,
+                    poisson_sampling=True,
+                    grad_sample_mode=self.GRAD_SAMPLE_MODE,
+                    opt_exclude_frozen=opt_exclude_frozen,
+                )
+                self._train_steps(model, optimizer, dl)
 
-    def _compare_to_vanilla(self, do_noise, do_clip, expected_match, use_closure=False):
+    def _compare_to_vanilla(
+        self,
+        do_noise,
+        do_clip,
+        expected_match,
+        grad_sample_mode,
+        use_closure=False,
+        max_steps=1,
+    ):
         torch.manual_seed(0)
         v_model, v_optimizer, v_dl = self._init_vanilla_training()
         if not use_closure:
-            self._train_steps(v_model, v_optimizer, v_dl, max_steps=1)
+            self._train_steps(v_model, v_optimizer, v_dl, max_steps=max_steps)
         else:
-            self._train_steps_with_closure(v_model, v_optimizer, v_dl, max_steps=1)
+            self._train_steps_with_closure(
+                v_model, v_optimizer, v_dl, max_steps=max_steps
+            )
         vanilla_params = [
             (name, p) for name, p in v_model.named_parameters() if p.requires_grad
         ]
@@ -197,12 +242,15 @@ class BasePrivacyEngineTest(ABC):
         p_model, p_optimizer, p_dl, _ = self._init_private_training(
             poisson_sampling=False,
             noise_multiplier=1.0 if do_noise else 0.0,
-            max_grad_norm=1.0 if do_clip else 99999.0,
+            max_grad_norm=0.1 if do_clip else 1e20,
+            grad_sample_mode=grad_sample_mode,
         )
         if not use_closure:
-            self._train_steps(p_model, p_optimizer, p_dl, max_steps=1)
+            self._train_steps(p_model, p_optimizer, p_dl, max_steps=max_steps)
         else:
-            self._train_steps_with_closure(p_model, p_optimizer, p_dl, max_steps=1)
+            self._train_steps_with_closure(
+                p_model, p_optimizer, p_dl, max_steps=max_steps
+            )
         private_params = [p for p in p_model.parameters() if p.requires_grad]
 
         for (name, vp), pp in zip(vanilla_params, private_params):
@@ -210,104 +258,69 @@ class BasePrivacyEngineTest(ABC):
                 # vanilla gradient is nearly zero: will match even with clipping
                 continue
 
+            atol = 1e-7 if max_steps == 1 else 1e-4
             self.assertEqual(
-                torch.allclose(vp, pp, atol=1e-8, rtol=1e-3),
+                torch.allclose(vp, pp, atol=atol, rtol=1e-3),
                 expected_match,
                 f"Unexpected private/vanilla weight match ({name})."
                 f"Should be: {expected_match}",
             )
             self.assertEqual(
-                torch.allclose(vp.grad, pp.grad, atol=1e-8, rtol=1e-3),
+                torch.allclose(vp.grad, pp.grad, atol=atol, rtol=1e-3),
                 expected_match,
                 f"Unexpected private/vanilla gradient match ({name})."
                 f"Should be: {expected_match}",
             )
 
-    def _compare_to_vanilla_accumulated(self, do_noise, do_clip, expected_match):
-        torch.manual_seed(0)
-        v_model, v_optimizer, v_dl = self._init_vanilla_training()
-        self._train_steps(v_model, v_optimizer, v_dl, max_steps=4)
-        vanilla_params = [
-            (name, p) for name, p in v_model.named_parameters() if p.requires_grad
-        ]
-
-        torch.manual_seed(0)
-        p_model, p_optimizer, p_dl, _ = self._init_private_training(
-            poisson_sampling=False,
-            noise_multiplier=1.0 if do_noise else 0.0,
-            max_grad_norm=10.0 if do_clip else 99999.0,
+    @given(
+        do_clip=st.booleans(),
+        do_noise=st.booleans(),
+        use_closure=st.booleans(),
+        max_steps=st.sampled_from([1, 4]),
+    )
+    @settings(deadline=None)
+    def test_compare_to_vanilla(
+        self,
+        do_clip: bool,
+        do_noise: bool,
+        use_closure: bool,
+        max_steps: int,
+    ):
+        """
+        Compare gradients and updated weights with vanilla model initialized
+        with the same seed
+        """
+        self._compare_to_vanilla(
+            do_noise=do_noise,
+            do_clip=do_clip,
+            expected_match=not (do_noise or do_clip),
+            use_closure=use_closure,
+            grad_sample_mode=self.GRAD_SAMPLE_MODE,
+            max_steps=max_steps,
         )
-        self._train_steps(p_model, p_optimizer, p_dl, max_steps=4)
-        private_params = [p for p in p_model.parameters() if p.requires_grad]
-
-        for (name, vp), pp in zip(vanilla_params, private_params):
-            if vp.grad.norm() < 1e-4:
-                # vanilla gradient is nearly zero: will match even with clipping
-                continue
-
-            self.assertEqual(
-                torch.allclose(vp, pp, atol=1e-8, rtol=1e-3),
-                expected_match,
-                f"Unexpected private/vanilla weight match ({name})."
-                f"Should be: {expected_match}",
-            )
-            self.assertEqual(
-                torch.allclose(vp.grad, pp.grad, atol=1e-8, rtol=1e-3),
-                expected_match,
-                f"Unexpected private/vanilla gradient match ({name})."
-                f"Should be: {expected_match}",
-            )
-
-    def test_compare_to_vanilla(self):
-        """
-        Compare gradients and updated weights with vanilla model initialized
-        with the same seed
-        """
-        for do_noise in (True, False):
-            for do_clip in (True, False):
-                for use_closure in (True, False):
-                    with self.subTest(
-                        do_noise=do_noise, do_clip=do_clip, use_closure=use_closure
-                    ):
-                        self._compare_to_vanilla(
-                            do_noise=do_noise,
-                            do_clip=do_clip,
-                            expected_match=not (do_noise or do_clip),
-                            use_closure=use_closure,
-                        )
-
-    def test_compare_to_vanilla_accumulated(self):
-        """
-        Compare gradients and updated weights with vanilla model initialized
-        with the same seed
-        """
-        for do_noise in (False, True):
-            for do_clip in (False, True):
-                with self.subTest(do_noise=do_noise, do_clip=do_clip):
-                    self._compare_to_vanilla(
-                        do_noise=do_noise,
-                        do_clip=do_clip,
-                        expected_match=not (do_noise or do_clip),
-                    )
 
     def test_flat_clipping(self):
         self.BATCH_SIZE = 1
         max_grad_norm = 0.5
 
+        torch.manual_seed(1337)
         model, optimizer, dl, _ = self._init_private_training(
             noise_multiplier=0.0,
             max_grad_norm=max_grad_norm,
             clipping="flat",
             poisson_sampling=False,
+            grad_sample_mode=self.GRAD_SAMPLE_MODE,
+        )
+        self._train_steps(model, optimizer, dl, max_steps=1)
+        clipped_grads = torch.cat(
+            [p.summed_grad.reshape(-1) for p in model.parameters() if p.requires_grad]
         )
-        optimizer.signal_skip_step()
 
+        torch.manual_seed(1337)
+        model, optimizer, dl = self._init_vanilla_training()
         self._train_steps(model, optimizer, dl, max_steps=1)
         non_clipped_grads = torch.cat(
-            [p.grad.view(-1) for p in model.parameters() if p.requires_grad]
-        )
-        clipped_grads = torch.cat(
-            [p.summed_grad.view(-1) for p in model.parameters() if p.requires_grad]
+            [p.grad.reshape(-1) for p in model.parameters() if p.requires_grad]
         )
 
         self.assertAlmostEqual(clipped_grads.norm().item(), max_grad_norm, places=3)
@@ -317,23 +330,27 @@ class BasePrivacyEngineTest(ABC):
         self.BATCH_SIZE = 1
         max_grad_norm_per_layer = 1.0
 
-        model, optimizer, dl, _ = self._init_private_training(
+        torch.manual_seed(1337)
+        p_model, p_optimizer, p_dl, _ = self._init_private_training(
             noise_multiplier=0.0,
             max_grad_norm=max_grad_norm_per_layer,
             clipping="per_layer",
             poisson_sampling=False,
+            grad_sample_mode=self.GRAD_SAMPLE_MODE,
         )
+        p_optimizer.signal_skip_step()
+        self._train_steps(p_model, p_optimizer, p_dl, max_steps=1)
 
-        optimizer.signal_skip_step()
+        torch.manual_seed(1337)
+        v_model, v_optimizer, v_dl = self._init_vanilla_training()
+        self._train_steps(v_model, v_optimizer, v_dl, max_steps=1)
 
-        self._train_steps(model, optimizer, dl, max_steps=1)
-
-        for p in model.parameters():
-            if not p.requires_grad:
+        for p_p, v_p in zip(p_model.parameters(), v_model.parameters()):
+            if not p_p.requires_grad:
                 continue
 
-            non_clipped_norm = p.grad.norm().item()
-            clipped_norm = p.summed_grad.norm().item()
+            non_clipped_norm = v_p.grad.norm().item()
+            clipped_norm = p_p.summed_grad.norm().item()
 
             self.assertAlmostEqual(
                 min(non_clipped_norm, max_grad_norm_per_layer), clipped_norm, places=3
@@ -347,6 +364,7 @@ class BasePrivacyEngineTest(ABC):
             poisson_sampling=False,
             noise_multiplier=0.0,
             max_grad_norm=99999.0,
+            grad_sample_mode=self.GRAD_SAMPLE_MODE,
         )
         self._train_steps(model, optimizer, dl, max_steps=1)
 
@@ -366,17 +384,31 @@ class BasePrivacyEngineTest(ABC):
         Test that adding noise results in ever different model params.
         We disable clipping in this test by setting it to a very high threshold.
         """
-        model, optimizer, dl, _ = self._init_private_training(poisson_sampling=False)
+        model, optimizer, dl, _ = self._init_private_training(
+            poisson_sampling=False, grad_sample_mode=self.GRAD_SAMPLE_MODE
+        )
         self._train_steps(model, optimizer, dl, max_steps=1)
         first_run_params = (p for p in model.parameters() if p.requires_grad)
 
-        model, optimizer, dl, _ = self._init_private_training(poisson_sampling=False)
+        model, optimizer, dl, _ = self._init_private_training(
+            poisson_sampling=False, grad_sample_mode=self.GRAD_SAMPLE_MODE
+        )
         self._train_steps(model, optimizer, dl, max_steps=1)
         second_run_params = (p for p in model.parameters() if p.requires_grad)
 
         for p0, p1 in zip(first_run_params, second_run_params):
             self.assertFalse(torch.allclose(p0, p1))
 
+    def test_get_compatible_module_inaction(self):
+        needs_no_replacement_module = nn.Linear(1, 2)
+        fixed_module = PrivacyEngine.get_compatible_module(needs_no_replacement_module)
+        self.assertFalse(fixed_module is needs_no_replacement_module)
+        self.assertTrue(
+            are_state_dict_equal(
+                needs_no_replacement_module.state_dict(), fixed_module.state_dict()
+            )
+        )
+
     def test_model_validator(self):
         """
         Test that the privacy engine raises errors
@@ -393,6 +425,7 @@ class BasePrivacyEngineTest(ABC):
                 data_loader=dl,
                 noise_multiplier=1.3,
                 max_grad_norm=1,
+                grad_sample_mode=self.GRAD_SAMPLE_MODE,
             )
 
     def test_model_validator_after_fix(self):
@@ -410,19 +443,10 @@ class BasePrivacyEngineTest(ABC):
             data_loader=dl,
             noise_multiplier=1.3,
             max_grad_norm=1,
+            grad_sample_mode=self.GRAD_SAMPLE_MODE,
         )
         self.assertTrue(1, 1)
 
-    def test_get_compatible_module_inaction(self):
-        needs_no_replacement_module = nn.Linear(1, 2)
-        fixed_module = PrivacyEngine.get_compatible_module(needs_no_replacement_module)
-        self.assertFalse(fixed_module is needs_no_replacement_module)
-        self.assertTrue(
-            are_state_dict_equal(
-                needs_no_replacement_module.state_dict(), fixed_module.state_dict()
-            )
-        )
-
     def test_make_private_with_epsilon(self):
         model, optimizer, dl = self._init_vanilla_training()
         target_eps = 2.0
@@ -439,6 +463,7 @@ class BasePrivacyEngineTest(ABC):
             target_delta=1e-5,
             epochs=epochs,
             max_grad_norm=1.0,
+            grad_sample_mode=self.GRAD_SAMPLE_MODE,
         )
         self._train_steps(model, optimizer, poisson_dl, max_steps=total_steps)
         self.assertAlmostEqual(
@@ -451,12 +476,16 @@ class BasePrivacyEngineTest(ABC):
         to produce same (deterministic) runs.
         """
         torch.manual_seed(0)
-        m1, opt1, dl1, _ = self._init_private_training()
+        m1, opt1, dl1, _ = self._init_private_training(
+            grad_sample_mode=self.GRAD_SAMPLE_MODE
+        )
         self._train_steps(m1, opt1, dl1)
         params1 = [p for p in m1.parameters() if p.requires_grad]
 
         torch.manual_seed(0)
-        m2, opt2, dl2, _ = self._init_private_training()
+        m2, opt2, dl2, _ = self._init_private_training(
+            grad_sample_mode=self.GRAD_SAMPLE_MODE
+        )
         self._train_steps(m2, opt2, dl2)
         params2 = [p for p in m2.parameters() if p.requires_grad]
 
@@ -466,8 +495,11 @@ class BasePrivacyEngineTest(ABC):
                 "Model parameters after deterministic run must match",
             )
 
-    def test_param_equal_module_optimizer(self):
-        """Test that the privacy engine raises error if nn.Module parameters are not equal to optimizer parameters"""
+    def test_validator_weight_update_check(self):
+        """
+        Test that the privacy engine raises error if ModuleValidator.fix(model) is
+        called after the optimizer is created
+        """
         model = models.densenet121(pretrained=True)
         num_ftrs = model.classifier.in_features
         model.classifier = nn.Sequential(nn.Linear(num_ftrs, 10), nn.Sigmoid())
@@ -486,6 +518,7 @@ class BasePrivacyEngineTest(ABC):
                 data_loader=dl,
                 noise_multiplier=1.1,
                 max_grad_norm=1.0,
+                grad_sample_mode=self.GRAD_SAMPLE_MODE,
             )
 
         # if optimizer is defined after ModuleValidator.fix() then raise no error
@@ -498,18 +531,48 @@ class BasePrivacyEngineTest(ABC):
             data_loader=dl,
             noise_multiplier=1.1,
             max_grad_norm=1.0,
+            grad_sample_mode=self.GRAD_SAMPLE_MODE,
         )
-        self.assertTrue(1, 1)
 
-    @given(noise_scheduler=st.sampled_from([None, StepNoise]))
+    def test_parameters_match(self):
+        dl = self._init_data()
+
+        m1 = self._init_model()
+        m2 = self._init_model()
+        m2.load_state_dict(m1.state_dict())
+        # optimizer is initialized with m2 parameters
+        opt = torch.optim.SGD(m2.parameters(), lr=0.1)
+
+        # the values are the identical
+        for p1, p2 in zip(m1.parameters(), m2.parameters()):
+            self.assertTrue(torch.allclose(p1, p2))
+
+        privacy_engine = PrivacyEngine()
+        # but model parameters and optimzier parameters must be the same object,
+        # not just same values
+        with self.assertRaises(ValueError):
+            privacy_engine.make_private(
+                module=m1,
+                optimizer=opt,
+                data_loader=dl,
+                noise_multiplier=1.1,
+                max_grad_norm=1.0,
+                grad_sample_mode=self.GRAD_SAMPLE_MODE,
+            )
+
+    @given(
+        noise_scheduler=st.sampled_from([None, StepNoise]),
+    )
     @settings(deadline=None)
-    def test_checkpoints(self, noise_scheduler: Optional[_NoiseScheduler]):
+    def test_checkpoints(self, noise_scheduler: Optional[Type[StepNoise]]):
         # 1. Disable poisson sampling to avoid randomness in data loading caused by changing seeds.
         # 2. Use noise_multiplier=0.0 to avoid randomness in torch.normal()
         # create a set of components: set 1
         torch.manual_seed(1)
         m1, opt1, dl1, pe1 = self._init_private_training(
-            noise_multiplier=0.0, poisson_sampling=False
+            noise_multiplier=0.0,
+            poisson_sampling=False,
+            grad_sample_mode=self.GRAD_SAMPLE_MODE,
         )
         s1 = (
             noise_scheduler(optimizer=opt1, step_size=1, gamma=1.0)
@@ -519,7 +582,9 @@ class BasePrivacyEngineTest(ABC):
         # create a different set of components: set 2
         torch.manual_seed(2)
         m2, opt2, _, pe2 = self._init_private_training(
-            noise_multiplier=2.0, poisson_sampling=False
+            noise_multiplier=2.0,
+            poisson_sampling=False,
+            grad_sample_mode=self.GRAD_SAMPLE_MODE,
         )
         s2 = (
             noise_scheduler(optimizer=opt2, step_size=1, gamma=2.0)
@@ -574,7 +639,9 @@ class BasePrivacyEngineTest(ABC):
         # recreate set 1 from scratch (set11) and check it is different from the trained set 2
         torch.manual_seed(1)
         m11, opt11, dl11, _ = self._init_private_training(
-            noise_multiplier=0.0, poisson_sampling=False
+            noise_multiplier=0.0,
+            poisson_sampling=False,
+            grad_sample_mode=self.GRAD_SAMPLE_MODE,
         )
         s11 = (
             noise_scheduler(optimizer=opt11, step_size=1, gamma=1.0)
@@ -603,7 +670,10 @@ class BasePrivacyEngineTest(ABC):
     )
     @settings(deadline=None)
     def test_noise_level(
-        self, noise_multiplier: float, max_steps: int, secure_mode: bool
+        self,
+        noise_multiplier: float,
+        max_steps: int,
+        secure_mode: bool,
     ):
         """
         Tests that the noise level is correctly set
@@ -617,6 +687,7 @@ class BasePrivacyEngineTest(ABC):
             model, optimizer, dl, _ = self._init_private_training(
                 noise_multiplier=noise_multiplier,
                 secure_mode=secure_mode,
+                grad_sample_mode=self.GRAD_SAMPLE_MODE,
             )
             for p in model.parameters():
                 p.data.zero_()
@@ -687,7 +758,7 @@ class SampleConvNet(nn.Module):
         self.conv2 = nn.Conv1d(16, 32, 3, 1)
         self.lnorm1 = nn.LayerNorm((32, 23))
         self.conv3 = nn.Conv1d(32, 32, 3, 1, bias=False)
-        self.instnorm1 = nn.InstanceNorm1d(32, affine=True)
+        # self.instnorm1 = nn.InstanceNorm1d(32, affine=True)
         self.convf = nn.Conv1d(32, 32, 1, 1)
         for p in self.convf.parameters():
             p.requires_grad = False
@@ -712,7 +783,7 @@ class SampleConvNet(nn.Module):
         x = self.lnorm1(x)  # -> [B, 32, 23]
         x = F.relu(x)  # -> [B, 32, 23]
         x = self.conv3(x)  # -> [B, 32, 21]
-        x = self.instnorm1(x)  # -> [B, 32, 21]
+        # x = self.instnorm1(x)  # -> [B, 32, 21]
         x = self.convf(x)  # -> [B, 32, 21]
         x = self.fc1(x)  # -> [B, 32, 17]
         x = self.lnorm2(x)  # -> [B, 32, 17]
@@ -734,14 +805,73 @@ class PrivacyEngineConvNetTest(BasePrivacyEngineTest, unittest.TestCase):
                 [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
             ),
         )
-        return DataLoader(ds, batch_size=self.BATCH_SIZE, drop_last=True)
+        return DataLoader(ds, batch_size=self.BATCH_SIZE, drop_last=False)
 
-    def _init_model(
-        self, private=False, state_dict=None, model=None, **privacy_engine_kwargs
-    ):
+    def _init_model(self):
         return SampleConvNet()
 
 
+class PrivacyEngineConvNetEmptyBatchTest(PrivacyEngineConvNetTest):
+    def setUp(self):
+        super().setUp()
+
+        # This will trigger multiple empty batches with poisson sampling enabled
+        self.BATCH_SIZE = 1
+
+    def test_checkpoints(self):
+        pass
+
+    def test_noise_level(self):
+        pass
+
+
+class PrivacyEngineConvNetFrozenTest(BasePrivacyEngineTest, unittest.TestCase):
+    def _init_data(self):
+        ds = FakeData(
+            size=self.DATA_SIZE,
+            image_size=(1, 35, 35),
+            num_classes=10,
+            transform=transforms.Compose(
+                [transforms.ToTensor(), transforms.Normalize((0.1307,), (0.3081,))]
+            ),
+        )
+        return DataLoader(ds, batch_size=self.BATCH_SIZE, drop_last=False)
+
+    def _init_model(self):
+        m = SampleConvNet()
+        for p in itertools.chain(m.conv1.parameters(), m.gnorm1.parameters()):
+            p.requires_grad = False
+
+        return m
+
+
+@unittest.skipIf(not _is_functorch_available(), "not supported in this torch version")
+class PrivacyEngineConvNetFrozenTestFunctorch(PrivacyEngineConvNetFrozenTest):
+    def setUp(self):
+        super().setUp()
+        self.GRAD_SAMPLE_MODE = "functorch"
+
+
+@unittest.skipIf(
+    torch.__version__ < API_CUTOFF_VERSION, "not supported in this torch version"
+)
+class PrivacyEngineConvNetTestExpandedWeights(PrivacyEngineConvNetTest):
+    def setUp(self):
+        super().setUp()
+        self.GRAD_SAMPLE_MODE = "ew"
+
+    @unittest.skip("Original p.grad is not available in ExpandedWeights")
+    def test_sample_grad_aggregation(self):
+        pass
+
+
+@unittest.skipIf(not _is_functorch_available(), "not supported in this torch version")
+class PrivacyEngineConvNetTestFunctorch(PrivacyEngineConvNetTest):
+    def setUp(self):
+        super().setUp()
+        self.GRAD_SAMPLE_MODE = "functorch"
+
+
 class SampleAttnNet(nn.Module):
     def __init__(self):
         super().__init__()
@@ -812,7 +942,7 @@ class PrivacyEngineTextTest(BasePrivacyEngineTest, unittest.TestCase):
             ds,
             batch_size=self.BATCH_SIZE,
             collate_fn=batch_second_collate,
-            drop_last=True,
+            drop_last=False,
         )
 
     def _init_model(
@@ -821,6 +951,13 @@ class PrivacyEngineTextTest(BasePrivacyEngineTest, unittest.TestCase):
         return SampleAttnNet()
 
 
+@unittest.skipIf(not _is_functorch_available(), "not supported in this torch version")
+class PrivacyEngineTextTestFunctorch(PrivacyEngineTextTest):
+    def setUp(self):
+        super().setUp()
+        self.GRAD_SAMPLE_MODE = "functorch"
+
+
 class SampleTiedWeights(nn.Module):
     def __init__(self, tie=True):
         super().__init__()
@@ -840,22 +977,59 @@ class SampleTiedWeights(nn.Module):
             self.fc2.weight = nn.Parameter(w.clone())
 
     def forward(self, x):
+        x = x.unsqueeze(1)
         x = self.emb(x)
         x = self.fc1(x)
         x = self.fc2(x)
+        x = x.squeeze(1)
 
         return x
 
 
 class PrivacyEngineTiedWeightsTest(BasePrivacyEngineTest, unittest.TestCase):
+    def setUp(self):
+        super().setUp()
+
     def _init_data(self):
         ds = TensorDataset(
             torch.randint(low=0, high=100, size=(self.DATA_SIZE,)),
             torch.randint(low=0, high=100, size=(self.DATA_SIZE,)),
         )
-        return DataLoader(ds, batch_size=self.BATCH_SIZE, drop_last=True)
+        return DataLoader(ds, batch_size=self.BATCH_SIZE, drop_last=False)
 
-    def _init_model(
-        self, private=False, state_dict=None, model=None, **privacy_engine_kwargs
-    ):
-        return SampleTiedWeights(tie=False)
+    def _init_model(self):
+        return SampleTiedWeights(tie=True)
+
+
+@unittest.skipIf(not _is_functorch_available(), "not supported in this torch version")
+class PrivacyEngineTiedWeightsTestFunctorch(PrivacyEngineTiedWeightsTest):
+    def setUp(self):
+        super().setUp()
+        self.GRAD_SAMPLE_MODE = "functorch"
+
+
+class ModelWithCustomLinear(nn.Module):
+    def __init__(self):
+        super().__init__()
+        self.fc1 = CustomLinearModule(4, 8)
+        self.fc2 = LinearWithExtraParam(8, 4)
+        self.extra_param = nn.Parameter(torch.randn(4, 4))
+
+    def forward(self, x):
+        x = self.fc1(x)
+        x = self.fc2(x)
+        x = x.matmul(self.extra_param)
+        return x
+
+
+@unittest.skipIf(not _is_functorch_available(), "not supported in this torch version")
+class PrivacyEngineCustomLayerTest(BasePrivacyEngineTest, unittest.TestCase):
+    def _init_data(self):
+        ds = TensorDataset(
+            torch.randn(self.DATA_SIZE, 4),
+            torch.randint(low=0, high=3, size=(self.DATA_SIZE,)),
+        )
+        return DataLoader(ds, batch_size=self.BATCH_SIZE, drop_last=False)
+
+    def _init_model(self):
+        return ModelWithCustomLinear()
